#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <SDL/SDL_ttf.h>

#include "irmp3io.h"

#define LEFTALIGN 1
#define CENTERALIGN 2
#define RIGHTALIGN 4

#define WITH_PREAMP_CODE 1

TTF_Font *font;
TTF_Font *smallfont;
char	bQuit = 0;

void RenderStatusText(SDL_Surface* surface, TTF_Font* font, const char texttorender[255], int xpos, int ypos, int width, int align, int anti, int blank);
void RenderStatusTextWithColor(SDL_Surface* surface, TTF_Font* font, const char texttorender[255], int xpos, int ypos, int width, int align, int anti, int blank, SDL_Color **fgbg);
void do_periodic(SDL_Surface *);
int	ReadConfig(char *, SDL_Surface *);
int PlaceBMP(char *name, int x, int y, SDL_Surface *surface);
int	keyloop(void *data);
int SendCommand(int iSocket, char *pCommand);

#define MAX_KEYS	512	// maximum number of a single keysym.

// config items
char	*keycommands[MAX_KEYS];
char	*server;
char	*defaultfont;
int		defaultfontsize;

int main (int argc, char *argv[])
{
	int iSocket;
	SDL_Surface *screen;
	SDL_Surface *statussurface;
	fd_set	fds;
	struct timeval	tv;
	int	i;

	/* Initialize the SDL library */
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
		exit(1);
	}

	/* Initialize the TTF library */
	if ( TTF_Init() < 0 ) {
			fprintf(stderr, "Couldn't initialize TTF: %s\n",SDL_GetError());
			exit(2);
	}
	atexit(SDL_Quit);

	screen = SDL_SetVideoMode(640, 480, 16, SDL_ANYFORMAT);
	if ( screen == NULL ) {
		fprintf(stderr, "Couldn't set 640x480x16 video mode: %s\n",
		SDL_GetError());
		exit(1);
	}

	//Blank the screen.
	SDL_FillRect(screen, NULL, SDL_MapRGBA(screen->format, 0, 0, 0, 0));

	statussurface = SDL_CreateRGBSurface(SDL_HWSURFACE, 640, 480, 16, 0, 0, 0, 0);

	// defaults for the configuration...
	server = "localhost";
	defaultfont = "/usr/share/fonts/default/TrueType/comic.ttf";	// works on RH 7.2
	defaultfontsize = 20;
	for (i=0; i< MAX_KEYS; i++)
		keycommands[i] = 0;

	if (!ReadConfig("irmp3_sdl_interface.conf", statussurface))
		exit(1);

	font = TTF_OpenFont(defaultfont, defaultfontsize);
	smallfont = TTF_OpenFont(defaultfont, 12);	// FIXME

#ifdef WITH_PREAMP_CODE
	PlaceBMP("MP3-off.bmp", 164, 20, statussurface);
	PlaceBMP("CD-off.bmp", 262, 20, statussurface);
	PlaceBMP("tuner-off.bmp", 336, 20, statussurface);
	PlaceBMP("video-off.bmp", 460, 20, statussurface);
#endif
	SDL_BlitSurface(statussurface, NULL, screen, NULL);
	SDL_Flip(screen);

	iSocket = irmp3_io_connect(server);
	if (iSocket > 0)
	{
		SDL_CreateThread(keyloop, &iSocket);
		while (bQuit == 0)
		{
			FD_ZERO(&fds);
			FD_SET(iSocket, &fds);
	
			tv.tv_sec = 0;
			tv.tv_usec = 500000; // 500mS
			if (select(iSocket + 1, &fds, 0, 0, &tv) == 0)
			{
//				do_periodic(statussurface);
//				SDL_BlitSurface(statussurface, NULL, screen, NULL);
//				SDL_Flip(screen);
			}
			if (FD_ISSET(iSocket, &fds))
			{
				irmp3_io_data(iSocket, statussurface);
				SDL_BlitSurface(statussurface, NULL, screen, NULL);
				SDL_Flip(screen);
			}
		}
		close (iSocket);
	}
	TTF_CloseFont(font);
	TTF_CloseFont(smallfont);
	SDL_Quit();
	return 1;
}

// fix this...
int SendCommand(int iSocket, char *pCommand)
{
	int iRetstat;
	if ((iRetstat = send(iSocket, pCommand, strlen(pCommand), 0)) >= 0)
		iRetstat = send(iSocket, "\r\n", 2, 0);
	return iRetstat;
}

int	keyloop(void *data)
{
	int	iSocket = *((int *)data);
	SDL_Event	event;
	while (bQuit == 0)
	{
		SDL_WaitEvent(&event);
		switch(event.type)
		{
			case SDL_QUIT:
				bQuit = 1;
				break;
			case SDL_KEYDOWN:
				if (event.key.keysym.sym < MAX_KEYS)
				{
					if (keycommands[event.key.keysym.sym])
					{
						SendCommand(iSocket, keycommands[event.key.keysym.sym]);
					}
				}
				break;
			default:
				break;
		}
	}
	return 1;
}

void do_periodic(SDL_Surface *surface)
{
	static int iPos;
	static char	direction = 3;

	SDL_Rect	pb_dark;
	SDL_Rect	pb_light;

	pb_dark.x = 0;
	pb_dark.y = 0;
	pb_dark.w = 640;
	pb_dark.h = 5;

	SDL_FillRect(surface, &pb_dark, SDL_MapRGBA(surface->format, 0, 50, 50, 0));

	pb_light.x = iPos;;
	pb_light.y = 0;
	pb_light.w = 25;
	pb_light.h = 5;

	SDL_FillRect(surface, &pb_light, SDL_MapRGBA(surface->format, 0, 150, 150, 0));
	iPos += direction;
	if (iPos >= 615)
	{
		direction = -3;
	}
	else if (iPos <= 0)
	{
		direction = 3;
	}
}

void irmp3_io_callback_artist(void *data, char *text)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	RenderStatusText(surface, font, text, 100, 100, 540, LEFTALIGN, 1, 1);
}

void irmp3_io_callback_album(void *data, char *text)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	RenderStatusText(surface, font, text, 100, 125, 540, LEFTALIGN, 1, 1);
}

void irmp3_io_callback_title(void *data, char *text)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	RenderStatusText(surface, font, text, 100, 150, 540, LEFTALIGN, 1, 1);
}

void irmp3_io_callback_year(void *data, char *text)
{
}

void irmp3_io_callback_comment(void *data, char *text)
{
}

void irmp3_io_callback_endofsong(void *data)
{
}

void irmp3_io_callback_file(void *data, char *text)
{
		SDL_Surface *surface = (SDL_Surface *)data;
	SDL_Color *fgbg[2];
	SDL_Color fg = {192,192,192,255};
	SDL_Color bg = {0,0,0,255};
	fgbg[0] = &fg;
	fgbg[1] = &bg;
	RenderStatusTextWithColor(surface, smallfont, text, 100, 200, 640, LEFTALIGN, 1, 1, fgbg);
}

void irmp3_io_callback_playlist(void *data, char *text)
{
}

void irmp3_io_callback_volume(void *data, int iVolume)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	char	caText[20];

	SDL_Rect	pb_dark;
	SDL_Rect	pb_light;

	pb_dark.x = 100;
	pb_dark.y = 455;
	pb_dark.w = (iVolume * 440) / 100;
	pb_dark.h = 25;

	SDL_FillRect(surface, &pb_dark, SDL_MapRGBA(surface->format, 0, 50, 50, 0));

	pb_light.x = pb_dark.w + 100;
	pb_light.y = 455;
	pb_light.w = 440 - pb_dark.w;
	pb_light.h = 25;

	SDL_FillRect(surface, &pb_light, SDL_MapRGBA(surface->format, 0, 150, 150, 0));

	sprintf(caText, "%d %%", iVolume);
	RenderStatusText(surface, font, caText, 320, 455, 640, CENTERALIGN, 1, 0);
}

void irmp3_io_callback_time(void *data, int amount_done, int amount_todo)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	char	caText[20];

	SDL_Rect	pb_dark;
	SDL_Rect	pb_light;

	if (amount_todo == 0 && amount_done == 0)	// prevent divide by zero.
		return;
	pb_dark.x = 100;
	pb_dark.y = 179;
	pb_dark.w = (amount_done * 440) / (amount_done + amount_todo);
	pb_dark.h = 25;

	SDL_FillRect(surface, &pb_dark, SDL_MapRGBA(surface->format, 0, 50, 50, 0));

	pb_light.x = pb_dark.w + 100;
	pb_light.y = 179;
	pb_light.w = 440 - pb_dark.w;
	pb_light.h = 25;

	SDL_FillRect(surface, &pb_light, SDL_MapRGBA(surface->format, 0, 150, 150, 0));

	sprintf(caText, "%d:%02d (%d:%02d)", amount_done / 60, amount_done % 60, amount_todo / 60, amount_todo % 60);
	RenderStatusText(surface, font, caText, 320, 175, 640, CENTERALIGN, 1, 0);
}

void irmp3_io_callback_genre(void *data, char *text)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	RenderStatusText(surface, font, text, 320, 215, 640, CENTERALIGN, 1, 1);
}

void irmp3_io_callback_beginofsong(void *data)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	PlaceBMP("play.bmp", 100, 20, surface);

	// artist
	RenderStatusText(surface, font, " ", 100, 100, 540, LEFTALIGN, 1, 1);

	// album
	RenderStatusText(surface, font, " ", 100, 125, 540, LEFTALIGN, 1, 1);

	// title
	RenderStatusText(surface, font, " ", 100, 150, 540, LEFTALIGN, 1, 1);

	// genre
	RenderStatusText(surface, font, " ", 320, 215, 640, CENTERALIGN, 1, 1);
}

void irmp3_io_callback_pause(void *data)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	PlaceBMP("pause.bmp", 100, 20, surface);
}

void irmp3_io_callback_stop(void *data)
{
	SDL_Surface *surface = (SDL_Surface *)data;
	PlaceBMP("stop.bmp", 100, 20, surface);
}


int PlaceBMP(char *name, int x, int y, SDL_Surface *surface)
{
	SDL_Surface	*icon;
	SDL_Rect	dest;
	icon = SDL_LoadBMP(name);
	if (!icon)
		return 0;
	dest.x = x;
	dest.y = y;
	dest.w = icon->w;
	dest.h = icon->h;
	
	SDL_BlitSurface(icon, NULL, surface, &dest);
	return 1;
}

void irmp3_io_callback_channel(void *data, char *text, int level)
{
#ifdef WITH_PREAMP_CODE
	SDL_Surface *surface = (SDL_Surface *)data;
	if (strcmp(text, "pcm") == 0)
	{
		if (level > 0)
			PlaceBMP("MP3-on.bmp", 164, 20, surface);
		else
			PlaceBMP("MP3-off.bmp", 164, 20, surface);
	}
	else if (strcmp(text, "CD") == 0)
	{
		if (level > 0)
			PlaceBMP("CD-on.bmp", 262, 20, surface);
		else
			PlaceBMP("CD-off.bmp", 262, 20, surface);
	}
	else if (strcmp(text, "line") == 0)
	{
		if (level > 0)
			PlaceBMP("tuner-on.bmp", 336, 20, surface);
		else
			PlaceBMP("tuner-off.bmp", 336, 20, surface);
	}
	else if (strcmp(text, "line1") == 0)
	{
		if (level > 0)
			PlaceBMP("video-on.bmp", 460, 20, surface);
		else
			PlaceBMP("video-off.bmp", 460, 20, surface);
	}
#endif
	return;

}

void irmp3_io_callback_mute(void *data, int status)
{
}

void RenderStatusText(SDL_Surface* surface, TTF_Font* font, const char texttorender[255], int xpos, int ypos, int width, int align, int anti, int blank)
{
	SDL_Color foregr = {255,255,255,255};
	SDL_Color backgr = {0,0,0,255};
	SDL_Color *fgbg[2];

	fgbg[0] = &foregr;
	fgbg[1] = &backgr;
	RenderStatusTextWithColor(surface, font, texttorender, xpos, ypos, width, align, anti, blank, fgbg);
}
void RenderStatusTextWithColor(SDL_Surface* surface, TTF_Font* font, const char texttorender[255], int xpos, int ypos, int width, int align, int anti, int blank, SDL_Color **fgbg)
{
	SDL_Color* fg = fgbg[0];
	SDL_Color* bg = fgbg[1];
	SDL_Rect back;
	SDL_Surface* text;

	if (anti == 0)
	{
		text = TTF_RenderText_Shaded(font, texttorender, *fg, *bg);
	}
	else
	{
		text = TTF_RenderText_Blended(font, texttorender, *fg);
	}
	
	if (text != NULL)
	{
		if (align == CENTERALIGN)
		{
			back.x = xpos - (width/2);
		}
		else if (align == RIGHTALIGN)
		{
			back.x = xpos - width;
		}
		else
		{
			back.x = xpos;
		}
		back.y = ypos;
		back.h = text->h;
		back.w = width;
		if (blank)
		{
			SDL_FillRect(surface, &back, SDL_MapRGBA(surface->format, 0, 0, 0, 0));
		}

		back.w = text->w;
		if (align == CENTERALIGN)
		{
			back.x = xpos - (text->w / 2);
		}
		else if (align == RIGHTALIGN)
		{
			back.x = xpos - text->w;
		}
		else
		{
			back.x = xpos;
		}

		SDL_BlitSurface(text, NULL, surface, &back);

		SDL_FreeSurface(text);
	}
}

int	ReadConfig(char *file, SDL_Surface *surface)
{
	char	caLine[512];
	char	*pArgs[20];
	int	iArgCount;
	int iLineCount = 0;
	FILE	*fp = fopen(file, "r");
	if (!fp)
	{
		perror(file);
		return 0;
	}
	while (fgets(caLine, sizeof(caLine), fp))
	{
		iLineCount++;
		ParseLine(caLine, pArgs, &iArgCount, 20);
		if (pArgs[0][0] == '#')
			continue;
		if (iArgCount > 0)
		{
			if (strcmp(pArgs[0], "key") == 0)
			{
				// FIXME:
				// for now you have to put the actual keycode number from /usr/include/SDL/SDK_keysym.h
				if (iArgCount >= 3)
				{
					int iKey = atoi(pArgs[1]);
					if (iKey > 0 && iKey < MAX_KEYS)
					{
						keycommands[iKey] = strdup(pArgs[2]);
//						printf("Key: %d, value '%s'\n", iKey, keycommands[iKey]);
					}
					else
						fprintf(stderr, "line %d: Keycode out of range\n", iLineCount);
				}
				else
					fprintf(stderr, "line %d: invalid Syntax\n", iLineCount);
			}
			else if (strcmp(pArgs[0], "server") == 0)
			{
				if (iArgCount >= 2)
					server = strdup(pArgs[1]);
				else
					fprintf(stderr, "line %d: invalid Syntax\n", iLineCount);
			}
			else if (strcmp(pArgs[0], "defaultfont") == 0)
			{
				if (iArgCount >= 3)
					defaultfontsize = atoi(pArgs[2]);
				if (iArgCount >= 2)
					defaultfont = strdup(pArgs[1]);
				else
					fprintf(stderr, "line %d: invalid Syntax\n", iLineCount);
			}
			else if (strcmp(pArgs[0], "label") == 0)
			{
				// syntax:
				// label x y "value" [align] [font] [size] [color r] [color b] [color g]
				int	x,y;
				int	align = LEFTALIGN;
				SDL_Color *fgbg[2];
				SDL_Color fg = {192,192,192,255};	// defaults...
				SDL_Color bg = {0,0,0,255};
				fgbg[0] = &fg;
				fgbg[1] = &bg;
				if (iArgCount >= 4) // minimum number of args
				{
					x = atoi(pArgs[1]);
					y = atoi(pArgs[2]);

					if (iArgCount >= 5)	// align
					{
						if (strcmp(pArgs[4], "left") == 0)	align = LEFTALIGN;
						else if (strcmp(pArgs[4], "right") == 0)	align = RIGHTALIGN;
						else if (strcmp(pArgs[4], "center") == 0)	align = CENTERALIGN;
						else
							fprintf(stderr, "line %d: invalid alignment\n", iLineCount);
					}
					if (iArgCount >= 6)	// font
					{
						if (strcmp(pArgs[5], "default") == 0)
							pArgs[5] = defaultfont;

						if (iArgCount >= 7)	// size
						{
							font = TTF_OpenFont(pArgs[5], atoi(pArgs[6]));
						}
						else
							font = TTF_OpenFont(pArgs[5], defaultfontsize);
					}
					else
						font = TTF_OpenFont(defaultfont, defaultfontsize);

					if (iArgCount >= 8) fg.r = atoi(pArgs[7]); // Red
					if (iArgCount >= 9) fg.g = atoi(pArgs[8]); // Green
					if (iArgCount >= 10) fg.b = atoi(pArgs[9]);	// Blue

					if (!font)
						fprintf(stderr, "line %d: Font Error\n", iLineCount);
					else
						RenderStatusTextWithColor(surface, font, pArgs[3], x, y, 640, align, 1, 1, fgbg);
				}
				else
					fprintf(stderr, "line %d: invalid Syntax\n", iLineCount);
			}
		}
	}
	fclose(fp);
	return 1;
}
